#include <iostream>
#include <cassert>
#include <cstdlib>
#include "ltlparse/public.hh"
#include "ltlvisit/relabel.hh"
#include "ltlvisit/tostring.hh"

using namespace spot::ltl;

void
syntax(char *prog)
{
  std::cerr << "syntax: " << prog << " formula\n";
  exit(2);
}

int main(int argc, char** argv)
{
  if (argc != 2)
    syntax(argv[0]);

  // Parse the input formula
  parse_error_list p;
  const formula* f = parse(argv[1], p);
  if (format_parse_errors(std::cerr, argv[1], p))
    return 2;

  // Relabel the formula using p0, p1, ...
  relabeling_map m;
  const formula* g = relabel(f, Pnn, &m);

  // Output the relabeled formula
  to_string(g, std::cout, true) << '\n';

  // Output the mapping of atomic propositions.
  for (relabeling_map::const_iterator i = m.begin(); i != m.end(); ++i)
    std::cout << "#define " << to_string(i->first)
	      << ' ' << to_spin_string(i->second, true) << '\n';

  g->destroy();
  f->destroy();
}
