#!/usr/bin/perl
# Runs experiment for given model, tool, and exactly one formula
# Expects: model file,
# 	   formula,
#          name of translator (+ translation options),
#          name of log file.


use strict;
use warnings;

use Switch;

use Getopt::Long qw(:config no_ignore_case);

my $verbose; my $help;
my $models_path = "../models";
my $model_name; my $formula; my $tool_name; my $tool_options=""; my $tool_path, my $log, my $por=1, my $tool_formula;
my $timeout='timeout -sKILL 30m'; my $timeout_ret = 35072;  
my $time; my $mem;

my $timeouted = 0;
my $ltlcross_res = 'ltlcross.csv';

### Automata statistics fields ###
my $aut_states; my $aut_edges; my $aut_trans; my $aut_nondet_states;
my $prod_states; my $prod_trans; my $prod_scc;
my $trim_aut_states; my $trim_aut_edges; my $trim_aut_trans;
my $trim_prod_states; my $trim_prod_trans;
my $UMrestr = "1";
my $lconly;

my $repetitions=5;
my $id;
  

GetOptions(
	   'm|model=s' => \$model_name,
	   'f|formula=s' => \$formula,
	   't|tool=s' => \$tool_name,
	   'o|options=s' => \$tool_options,
	   'v|verbose' => \$verbose,
	   'l|log=s' => \$log,
	   'h|help' => \$help,
	   'p|por!' => \$por,
	   'um=s' => \$UMrestr,
	   'ltlcross_only' => \$lconly,
	   'n|repetitions=i' => \$repetitions,
	   'id=s' => \$id,
	   );

### Runs ltlcross and stores statistics in $ltlcross_res file ###
# TODO move into one switch with others, substitute options here by variable $tool_options etc...
sub ltlcross {
  my $ltlcross_conf;
  switch ($tool_name) {
  case /^Spot$/ { 
    $ltlcross_conf = 'ltl2tgba -s %s >%N';
    }
  case /^Spin$/ {
    $ltlcross_conf = 'spin -f %s >%N';
    }
  case /^noJumpSpot$/ {
    $ltlcross_conf = 'ltl2tgba -x degen-lskip=0 -s %s >%N';
    }
  case /^detSpot$/ {
    $ltlcross_conf = 'ltl2tgba -s --deterministic %s >%N';
  }
  case /LTL2BA/ {
    $ltlcross_conf = 'ltl2ba -f %s >%N';
    }
  case /^LTL3BA$/ {
    $ltlcross_conf = 'ltl3ba -f %s -S >%N';
  }
  case /detLTL3BA/ {
    $ltlcross_conf = 'ltl3ba -f %s -S -M >%N';
  }
  case /Modella/ {
    $ltlcross_conf = 'mod2spin -f %s | sed -e "s/\(:: \)\([^\(]*\)\( -> goto .*\)/\1\(\2\)\3/g">%N';
  }
  }
  my $ltlcross_com_UMP = "ltlcross --csv=$ltlcross_res --universal='".$UMrestr."' --no-check -f '$formula' '$ltlcross_conf'";
  my $ltlcross_com = "ltlcross --csv=$ltlcross_res --products=0 -f '$formula' '$ltlcross_conf'";
  print $ltlcross_com_UMP."\n" if ($verbose);
  `timeout -sKILL 5m $ltlcross_com_UMP >/dev/null 2>/dev/null || $ltlcross_com >/dev/null 2>/dev/null`;
}

### parses $ltlcross_res file and gathers statistics abou generatedt automaton ###
sub parse_ltlcross {
  my @ltlcross_data;
  open(my $fh, '<', $ltlcross_res) or die "Can't read file '$ltlcross_res' [$!]\n";
  while (my $line = <$fh>) {
    chomp $line;
    my @fields = split(/,/, $line);
    push @ltlcross_data, \@fields;
  }
  my %ltlcross;
  my $names = $ltlcross_data[0];
  for (my $i = 0; $i < scalar @$names; $i++) {
    $ltlcross{$ltlcross_data[0][$i]}=$ltlcross_data[1][$i];
  }
  $aut_states = $ltlcross{'"states"'};
  $aut_edges = $ltlcross{'"edges"'};
  $aut_trans = $ltlcross{'"transitions"'};
  $aut_nondet_states = $ltlcross{'"nondet_states"'};
  $prod_states = $ltlcross{'"product_states"'};
  $prod_trans = $ltlcross{'"product_transitions"'};
  $prod_scc = $ltlcross{'"product_scc"'};
  $trim_aut_states = $ltlcross{'"trim_states"'};
  $trim_aut_edges = $ltlcross{'"trim_edges"'};
  $trim_aut_trans = $ltlcross{'"trim_transitions"'};
  $trim_prod_states = $ltlcross{'"product_trim_states"'};
  $trim_prod_trans = $ltlcross{'"product_trim_transitions"'};
}

sub parse_time {
  die ("cannot open $time") unless open (TIME,"time");
  if (<TIME> =~ /mem: (\d+), time: (\d+.\d\d)\+(\d+.\d\d)/) {
    $mem = $1;
    $time = $2+$3;
  }
  close TIME;
}

if ($help) {
# TODO HELP
print<<EOF;
"Model, formula, tool, log options are required!"  
EOF
exit 0;
}

unless ($model_name and $formula and $tool_name and $log) {
  print "Model, formula, tool, log options are required!\n";
  exit 1;
}

### Preparing tmp directory ###
  $log =~ /log(\d+)\.csv/;
  my $log_n = $1;
  print "mkdir tmp$log_n\ncd tmp$log_n\n" if $verbose;
  system "mkdir tmp$log_n";
  chdir "tmp$log_n";

### Relabel formula using apcollect, stores the list of defines into file "never" ###
`echo \"$formula\" > formula`;
print "Relabel AP in formula and generating of list of #defines...\n",
  "../apcollect \'$formula\'\n" if $verbose;
open(LABEL, "-|", "../apcollect \'$formula\'");
$formula = <LABEL>;
chomp $formula;
open(NEVER, ">never");
while(<LABEL>) {
  print NEVER $_;
}
print NEVER "\n";
close NEVER;
  
### Generating of neverclaim ###
print "Running ltlfilt...\n" if $verbose;
switch ($tool_name) {
  case /^Spot$/ { 
    $tool_formula = `ltlfilt -r0 -f \'$formula\' --format=\'%f\'`;
    $tool_path = "ltl2tgba";
    $tool_options.= " -s";
    }
  case /^Spin$/ {
    $tool_formula = `ltlfilt -r0 -f \'$formula\' -s --format=\'%f\'`;
    $tool_path = "spin";
#    $tool_options.= "";
    }
  case /^noJumpSpot$/ { 
    $tool_formula = `ltlfilt -r0 -f \'$formula\' --format=\'%f\'`;
    $tool_path = "ltl2tgba";
    $tool_options.= " -x degen-lskip=0 -s";
    }
  case /^detSpot$/ {
    $tool_formula = `ltlfilt -r0 -f \'$formula\' --format=\'%f\'`;
    $tool_path = "ltl2tgba";
    $tool_options.= " -s --deterministic";
  }
  case /LTL2BA/ {
    $tool_formula = `ltlfilt -r0 -f \'$formula\' -s --format=\'%f\'`;
    $tool_path = "ltl2ba";
    }
  case /^LTL3BA$/ {
    $tool_formula = `ltlfilt -r0 -f \'$formula\' -s --format=\'%f\'`;
    $tool_path = "ltl3ba";
    $tool_options = " -S";
  }
  case /detLTL3BA/ {
    $tool_formula = `ltlfilt -r0 -f \'$formula\' -s --format=\'%f\'`;
    $tool_path = "ltl3ba";
    $tool_options = " -S -M";
  }
  case /Modella/ {
    $tool_formula = `ltlfilt -r0 -f \'$formula\' -s --format=\'%f\'`;
    $tool_path = "mod2spin";
  }
}

ltlcross;
parse_ltlcross;

  chomp $tool_formula;
  print "$tool_formula\n" if $verbose;
  
  my $command = "$tool_path $tool_options -f \'$tool_formula\'";
  print "$command >> never\n" if $verbose;
  `$command >> never`;
  
  print "Result of the translation:\n" if $verbose;
  print `cat never` if $verbose; 
  
### Copies the model into tmp directory ###
  `cp $models_path/$model_name .`;
  $model_name =~ s/(.*\/)*(.*)$/$2/;
#  print `ls`;
  
  
### Compile the verifier, run it and store result into file ###
my $time_comm = "/usr/bin/time -o time --format \"mem: %M, time: %S+%U\"";
  print "$time_comm $timeout spin -a -N never $model_name > spin.log\n" if $verbose;
  `$time_comm $timeout spin -a -N never $model_name > spin.log`;
  `mv time stime`;
  `gcc -DMEMLIM=20480 -o pan pan.c` if $por;
  `gcc -DMEMLIM=20480 -DNOREDUCE -o pan pan.c` unless $por;
sub run {
  print "$time_comm $timeout ./pan -a -m100000000 > res" if $verbose;
  `$time_comm $timeout ./pan -a -m100000000 > res`;
  if ($? == $timeout_ret) {
    $timeouted = $?;
  }
}

### Parse the res file and extract desired values ###
  my $stored; my $visited; my $matched; my $trans; my $err; my $depth;
  my $m1; my $m2; my $m3; my $m4; my $memory; my $pan_time;

sub parse {
unless($timeouted) {
  open(RESULT,"res") or die ("Cannot open the file with results");
  while (<RESULT>) {
    $stored = $1 if /^\s*((\d+\.)?\d+(e\+\d+)?)\sstates, stored/;
    $visited = $1 if /\(((\d+\.)?\d+(e\+\d+)?) visited\)/;
    $matched = $1 if /^\s*((\d+\.)?\d+(e\+\d+)?)\sstates, matched/;
    $trans = $1 if /^\s*((\d+\.)?\d+(e\+\d+)?)\stransitions/;# \(\= stored\+matched\)/;
    
    $depth = $1 if /depth reached (\d+), errors: (\d+)/;
    $err = $2 if /depth reached (\d+), errors: (\d+)/;
    
    $m1 = $1 if /(\d+\.\d+)\tequivalent memory usage for states/;
    $m2 = $1 if /(\d+\.\d+)\tactual memory usage for states/;
    $m3 = $1 if /(\d+\.\d+)\tmemory used for hash table/;
    $m4 = $1 if /(\d+\.\d+)\tmemory used for DFS stack/;
    $memory = $1 if /(\d+\.\d+)\ttotal actual memory usage/;
    $pan_time = $1 if /pan: elapsed time (\d+(\.\d+)?(e\+\d+)?) seconds/;
  }
  $visited = $stored unless $visited;
  close RESULT;
  parse_time;
  }
}
my @times;
my $total_time=0;
for (my $i=1;$i <= $repetitions;$i++) {
  run unless ($lconly);
  parse unless ($lconly);
  if (!defined($pan_time)) {
    $i=5;
    next;
  }
  $total_time+=$pan_time;
  $times[$i-1]=$pan_time;
}
$pan_time=$total_time/$repetitions;
my @sorted_times = sort { $a <=> $b } @times;
my $min_time = $sorted_times[0];
my $median = $sorted_times[($repetitions-1)/2];
  
sub store {
### Prints extracted values into log file in csv format ###
  chdir "..";
#  if($stored) {`rm -r tmp$log_n`};
  open(LOG,">$log");
#  print LOG "model;formula;translator;aut_states;aut_nondet_states;aut_edges;aut_trans;stored;matched;visited;trans;depth;error;time;memory;Spin_time;timeout;POR;translation command;equivalent memory usage for states;actual memory usage for states;memory used for hash table;memory used for DFS stack;total actual memory usage'
unless ($timeouted) {
  print LOG "$id,$model_name,$formula,$tool_name,$aut_states,$aut_nondet_states,$aut_edges,$aut_trans,$trim_aut_states,$trim_aut_edges,$trim_aut_trans,$stored,$visited,$matched,$trans,$depth,$err,$time,$mem,$pan_time,$min_time,$median,0,1,$command,$prod_states,$prod_trans,$trim_prod_states,$trim_prod_trans,$prod_scc,$m1,$m2,$m3,$m4,$memory\n" if $por;
  print LOG "$id,$model_name,$formula,$tool_name,$aut_states,$aut_nondet_states,$aut_edges,$aut_trans,$trim_aut_states,$trim_aut_edges,$trim_aut_trans,$stored,$visited,$matched,$trans,$depth,$err,$time,$mem,$pan_time,$min_time,$median,0,0,$command,$prod_states,$prod_trans,$trim_prod_states,$trim_prod_trans,$prod_scc,$m1,$m2,$m3,$m4,$memory\n" unless $por;
}
if ($timeouted) {
  print LOG "$id,$model_name,$formula,$tool_name,$aut_states,$aut_nondet_states,$aut_edges,$aut_trans,$trim_aut_states,$trim_aut_edges,$trim_aut_trans,,,,,,,,,,,,$timeouted,1,$command,$prod_states,$prod_trans,$trim_prod_states,$trim_prod_trans,$prod_scc,,,,,\n" if $por;
  print LOG "$id,$model_name,$formula,$tool_name,$aut_states,$aut_nondet_states,$aut_edges,$aut_trans,$trim_aut_states,$trim_aut_edges,$trim_aut_trans,,,,,,,,,,,,$timeouted,0,$command,$prod_states,$prod_trans,$trim_prod_states,$trim_prod_trans,$prod_scc,,,,,\n" unless $por;
}
  close LOG;
}
store;
#  print "Removing the neverclaim...\n";
#  `rm never`;

exit 0;
